#include <cstdio>
#include <sys/time.h> /* gettimeofday, setitimer */
#include <termio.h>   /* termio */

#include <csignal>

#include <algorithm>
#include <list>

/* Captures video from a X window.
 * Requires the SHM extension to work on tolerable speeds.
 *
 * Copyright (C) 1992,2004 Bisqwit (http://iki.fi/bisqwit/)
 *
 * TODO: Create a standalone program without all parameters being
 *       hardcoded into the program.
 *
 * To compile: gcc -o xgetimage xgetimage.c -L/usr/X11R6/lib -lX11 -lXext -lXmu
 *
 * To postprocess the results, example:
 *  riffdump -nmovi -mfps=17.94 test0.avi
 *  riffdump -nmovi -mfps=17.94 test1.avi
 *  cat test{0,1}.avi | mencoder -forceidx - -ss 1 \
 *      -oac copy -audiofile ~/wine-audiolog.mp3 \
 *      -ovc copy -o xgetimage/test-short.avi
 */

#include "settings.hh"
#include "buffer.hh"
#include "msg.hh"
#include "x.hh"

typedef unsigned long long UINT64;
static UINT64 getutime(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (tv.tv_sec * 1000000LL) + tv.tv_usec;
}

static unsigned frames_total    = 0;
static unsigned frames_thisfile = 0;
static UINT64 start_time;

static std::list<FILE *> finished_encoders;
static bool empty = false;
static FILE *avifp= NULL;

static FILE *OpenAvi()
{
    static unsigned avinumber = 0;
    char CmdLine[512];
    char avifn[64];
    sprintf(avifn, "test%d.avi", avinumber++);

#if USE_MENCODER
    sprintf(CmdLine,
        "animmaker -o- -f%g -"
        "|mencoder - -o %s -ovc lavc -lavcopts"
        " vbitrate=%u:keyint=%d:qpel:dia=-3 >& /dev/null",
            (double)FPS,
            avifn,
            BITRATE,
            (int)(FPS * 3));
#else
    sprintf(CmdLine,
        "animmaker -f%g -o%s -",
            (double)FPS,
            avifn);
#endif

    MessageCreatingFile(avifn);

    frames_thisfile = 0;
    return popen(CmdLine, "w");
}

static void InitAvi()
{
    start_time   = getutime();
    avifp        = OpenAvi();
    frames_total = 0;
    empty        = true;
}

void BufferEmptied(FILE *fp)
{
    //fprintf(stderr, "\nBuffer %X emptied\n", fp);
    for(std::list<FILE*>::iterator
        j, i = finished_encoders.begin();
           i != finished_encoders.end();
        i = j)
    {
        j = i; ++j;
        
        if(fp == *i)
        {
            //fprintf(stderr, "\nFound and closed.\n");
            pclose(*i);
            finished_encoders.erase(i);
            return;
        }
    }
    if(fp == avifp) empty = true; 
}

static void DoneAvi()
{
    FlushBuffers();
    
    for(std::list<FILE*>::iterator
        i = finished_encoders.begin();
        i != finished_encoders.end();
        ++i)
    {
        pclose(*i);
    }
    
    finished_encoders.clear();
    avifp   = NULL;
    empty   = true;
}
static void NextAvi()
{
    if(empty)
        pclose(avifp);
    else
        finished_encoders.push_back(avifp);
    avifp = OpenAvi();
    empty = true;
}

static double GetVideoExcess()
{
    UINT64 time_now = getutime();
    double videopos = frames_total / (double)FPS;
    double want_pos = (time_now - start_time) / 1e6;
    return videopos - want_pos;
}

static void SendImage(unsigned char *vidbuf)
{
    static const unsigned char tgaBuf[] = {0,0,2, 0,0, 0,0, 0,  0,0, 0,0, 0,0, 0,0, 16,32};
    unsigned framesize = sizeof(tgaBuf) + width*height*2;
    unsigned opos      = sizeof(tgaBuf);
    unsigned frames_now;
#if 0
    unsigned y;
#endif
    
    BufferType thisframe(framesize);
    
    std::copy(tgaBuf, tgaBuf + sizeof tgaBuf,
              &thisframe[0]);
    
    thisframe[12] = width&255;
    thisframe[13] = width>>8; 
    thisframe[14] = height&255;
    thisframe[15] = height>>8; 

#if 0
    for(y=height; y-->0; )
    {
        unsigned x, ipos = y*width*2;
        for(x=0; x<width; ++x)
        {
            register unsigned short pixel = vidbuf[ipos] + (vidbuf[ipos+1]<<8);

#define   RED_LOW_BIT_MASK 0x0800
#define GREEN_LOW_BIT_MASK 0x0020
#define  BLUE_LOW_BIT_MASK 0x0001
#define MAX_RED   0x1F
#define MAX_GREEN 0x3F
#define MAX_BLUE  0x1F

register unsigned char R = (((pixel/RED_LOW_BIT_MASK  ) & MAX_RED  ) * (0x20 / (MAX_RED  +1)));
register unsigned char G = (((pixel/GREEN_LOW_BIT_MASK) & MAX_GREEN) * (0x40 / (MAX_GREEN+1)));
register unsigned char B = (((pixel/BLUE_LOW_BIT_MASK ) & MAX_BLUE ) * (0x20 / (MAX_BLUE +1)));
                                                                                              
            thisframe[opos++] = (B | (G << 5)) & 255;                                         
            thisframe[opos++] = (G >> 3) | (R << 3); 

            ipos += 2;
        }
    }
#else
    std::copy(vidbuf, vidbuf + width*height*2,
              &thisframe[opos]);
#endif
    
    frames_now = 0;
    
    // Set a high threshold for the first frame
    // to make it unlikely to discard frames
    double threshold = 0.333 / FPS;
    
    for(;;)
    {
        double excess = GetVideoExcess();
        if(excess > threshold) break; /* discard frame, we've had enough already */
        threshold = 0;

        if(frames_thisfile >= FRAMES_PER_FILE)
        {
            NextAvi();
        }
        
        empty = false;
        AddWriteTask(avifp, thisframe);
        
        ++frames_total;
        ++frames_thisfile;
        ++frames_now;
    }
    
    if(!frames_now)
        MessageDiscardFrame();
    else
        MessageCapturedFrame(frames_now);
}

static termio back;
static void InitKB()
{
    struct termio term;
    ioctl(0, TCGETA, &term);
    ioctl(0, TCGETA, &back);
    term.c_lflag &= ~(ECHO | ICANON);
    term.c_cc[VMIN] = 0; 
    ioctl(0, TCSETA, &term);
}
static void DoneKB()
{
    ioctl(0, TCSETA, &back);
}

static void SetTimeVal(struct timeval& t, double value)
{
    t.tv_sec  = (unsigned) value;
    t.tv_usec = (unsigned) ((value - t.tv_sec) * 1e6);
}

static void CaptureFrame(int) { /* nothing to do */ }
static void InitializeTimers()
{
    struct itimerval tr;
    SetTimeVal(tr.it_interval, 1.0 / FPS);
    
    double next = GetVideoExcess() * 0.98;

    // Set a minimum to ensure the timer won't be disabled
    if(next <= 1.4e-6) next = 1.4e-6;

    SetTimeVal(tr.it_value,    next);

    setitimer(ITIMER_REAL, &tr, NULL);
    
    /*
    *   FIXME: SEE ALSO "man 3p timer_create", "man 3p timer_getoverrun"
    *
    **/
}
static void EndTimers()
{
    struct itimerval tr;
    SetTimeVal(tr.it_value,    0);
    SetTimeVal(tr.it_interval, 0);
    setitimer(ITIMER_REAL, &tr, NULL);
    signal(SIGALRM, SIG_DFL);
}

int main(void)
{
    XInit();
    
    // Get 1 image to ensure there are no
    // first-initialization-delays waiting for us.
    GetImage();
    DeleteImage();
    
    InitKB();
    
    signal(SIGALRM, CaptureFrame);

    InitAvi();
    
    InitializeTimers();
    
    while(!BufferWaitInput())
    {
        unsigned char *data = GetImage();
        if(data)
        {
            SendImage(data);
            MessageFrameCounter(frames_total);
        }
        DeleteImage();
        
        InitializeTimers();
    }
    
    EndTimers();
    
    DoneKB();
    DoneAvi();
    
    MessageEnd();
    
    XDone();
    
    return 0;
}
